@@ -122,6 +122,7 @@ typeOfClick (Window w, XEvent * ev, gboolean allow_double_click)
         total += 10;
         if (XCheckMaskEvent (dpy, ButtonReleaseMask | ButtonPressMask, ev))
         {
+            last_timestamp = stashEventTime (last_timestamp, ev);
             if (ev->xbutton.button == button)
             {
                 clicks++;
@@ -135,6 +136,7 @@ typeOfClick (Window w, XEvent * ev, gboolean allow_double_click)
         }
         if (XCheckMaskEvent (dpy, ButtonMotionMask | PointerMotionMask, ev))
         {
+            last_timestamp = stashEventTime (last_timestamp, ev);
             xcurrent = ev->xmotion.x_root;
             ycurrent = ev->xmotion.y_root;
         }
@@ -157,7 +159,7 @@ clear_timeout (void)
 static inline gboolean
 raise_cb (gpointer data)
 {
-    Client *c;
+    Client *c = NULL;
     TRACE ("entering raise_cb");
 
     clear_timeout ();
@@ -165,6 +167,7 @@ raise_cb (gpointer data)
     if (c)
     {
         clientRaise (c);
+        clientPassGrabButton1 (c);
     }
     return (TRUE);
 }
@@ -193,7 +196,7 @@ moveRequest (Client * c, XEvent * ev)
 static inline void
 resizeRequest (Client * c, int corner, XEvent * ev)
 {
-    clientSetFocus (c, TRUE, FALSE);
+    clientSetFocus (c, FOCUS_SORT);
 
     if (CLIENT_FLAG_TEST_ALL (c,
             CLIENT_FLAG_HAS_RESIZE | CLIENT_FLAG_IS_RESIZABLE))
@@ -261,7 +264,9 @@ handleMotionNotify (XMotionEvent * ev)
                 workspaceSwitch (workspace + 1, NULL);
             }
             while (XCheckWindowEvent(dpy, ev->window, PointerMotionMask, (XEvent *) ev))
-                ; /* VOID */
+            {
+                last_timestamp = stashEventTime (last_timestamp, (XEvent *) ev);
+            }    
         }
     }
 }
@@ -269,7 +274,7 @@ handleMotionNotify (XMotionEvent * ev)
 static inline void
 handleKeyPress (XKeyEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
     int state, key;
 
     TRACE ("entering handleKeyEvent");
@@ -336,9 +341,11 @@ handleKeyPress (XKeyEvent * ev)
                 break;
             case KEY_RAISE_WINDOW:
                 clientRaise (c);
+                clientPassGrabButton1 (NULL);
                 break;
             case KEY_LOWER_WINDOW:
                 clientLower (c);
+                clientPassGrabButton1 (NULL);
                 break;
             case KEY_MOVE_NEXT_WORKSPACE:
                 workspaceSwitch (workspace + 1, c);
@@ -484,6 +491,7 @@ edgeButton (Client * c, int part, XButtonEvent * ev)
         if (tclick == XFWM_BUTTON_CLICK)
         {
             clientLower (c);
+            clientPassGrabButton1 (NULL);
         }
         else
         {
@@ -495,6 +503,7 @@ edgeButton (Client * c, int part, XButtonEvent * ev)
         if (ev->button == Button1)
         {
             clientRaise (c);
+            clientPassGrabButton1 (c);
         }
         if ((ev->button == Button1) || (ev->button == Button3))
         {
@@ -512,8 +521,9 @@ button1Action (Client * c, XButtonEvent * ev)
     g_return_if_fail (c != NULL);
     g_return_if_fail (ev != NULL);
 
-    clientSetFocus (c, TRUE, FALSE);
+    clientSetFocus (c, FOCUS_SORT);
     clientRaise (c);
+    clientPassGrabButton1 (c);
 
     tclick = typeOfClick (c->frame, &copy_event, TRUE);
 
@@ -555,6 +565,7 @@ titleButton (Client * c, int state, XButtonEvent * ev)
     else if (ev->button == Button2)
     {
         clientLower (c);
+        clientPassGrabButton1 (NULL);
     }
     else if (ev->button == Button3)
     {
@@ -574,10 +585,11 @@ titleButton (Client * c, int state, XButtonEvent * ev)
         }
         else
         {
-            clientSetFocus (c, TRUE, FALSE);
+            clientSetFocus (c, FOCUS_SORT);
             if (params.raise_on_click)
             {
                 clientRaise (c);
+                clientPassGrabButton1 (c);
             }
             ev->window = ev->root;
             if (button_handler_id)
@@ -618,6 +630,7 @@ rootScrollButton (XButtonEvent * ev)
 
     while (XCheckTypedWindowEvent (dpy, root, ButtonPress, &otherEvent))
     {
+        last_timestamp = stashEventTime (last_timestamp, &otherEvent);
         if (otherEvent.xbutton.button != ev->button)
         {
             XPutBackEvent (dpy, &otherEvent);
@@ -643,7 +656,7 @@ rootScrollButton (XButtonEvent * ev)
 static inline void
 handleButtonPress (XButtonEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
     Window win;
     int state, replay = FALSE;
 
@@ -665,6 +678,7 @@ handleButtonPress (XButtonEvent * ev)
         else if ((ev->button == Button2) && (state == AltMask) && (params.easy_click))
         {
             clientLower (c);
+            clientPassGrabButton1 (NULL);
         }
         else if ((ev->button == Button3) && (state == AltMask) && (params.easy_click))
         {
@@ -689,10 +703,11 @@ handleButtonPress (XButtonEvent * ev)
         {
             if (ev->button <= Button3)
             {
-                clientSetFocus (c, TRUE, FALSE);
+                clientSetFocus (c, FOCUS_SORT);
                 if (params.raise_on_click)
                 {
                     clientRaise (c);
+                    clientPassGrabButton1 (c);
                 }
                 clientButtonPress (c, win, ev);
             }
@@ -721,10 +736,11 @@ handleButtonPress (XButtonEvent * ev)
                 }
                 else
                 {
-                    clientSetFocus (c, TRUE, FALSE);
+                    clientSetFocus (c, FOCUS_SORT);
                     if (params.raise_on_click)
                     {
                         clientRaise (c);
+                        clientPassGrabButton1 (c);
                     }
                     ev->window = ev->root;
                     if (button_handler_id)
@@ -745,6 +761,7 @@ handleButtonPress (XButtonEvent * ev)
                 && (state == (AltMask | ControlMask))))
         {
             clientLower (c);
+            clientPassGrabButton1 (NULL);
         }
         else if ((win == MYWINDOW_XWINDOW (c->corners[CORNER_TOP_LEFT]))
             && (state == 0))
@@ -785,10 +802,15 @@ handleButtonPress (XButtonEvent * ev)
         {
             if (ev->button == Button1)
             {
-                clientSetFocus (c, TRUE, FALSE);
+                if (ev->window == c->window)
+                {
+                    clientPassGrabButton1 (c);
+                }
+                clientSetFocus (c, FOCUS_SORT);
                 if ((params.raise_on_click) || !CLIENT_FLAG_TEST (c, CLIENT_FLAG_HAS_BORDER))
                 {
                     clientRaise (c);
+                    clientPassGrabButton1 (c);
                 }
             }
             if (ev->window == c->window)
@@ -830,7 +852,7 @@ handleButtonRelease (XButtonEvent * ev)
 static inline void
 handleDestroyNotify (XDestroyWindowEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleDestroyNotify");
     TRACE ("DestroyNotify on window (0x%lx)", ev->window);
@@ -854,7 +876,7 @@ handleDestroyNotify (XDestroyWindowEvent * ev)
 static inline void
 handleMapRequest (XMapRequestEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleMapRequest");
     TRACE ("MapRequest on window (0x%lx)", ev->window);
@@ -891,7 +913,7 @@ handleMapRequest (XMapRequestEvent * ev)
 static inline void
 handleMapNotify (XMapEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleMapNotify");
     TRACE ("MapNotify on window (0x%lx)", ev->window);
@@ -917,13 +939,13 @@ handleUnmapNotify (XUnmapEvent * ev)
     
     if (ev->from_configure)
     {
-        TRACE ("Ignoring UnmapNotify caused by parent's resize\n");
+        TRACE ("Ignoring UnmapNotify caused by parent's resize");
         return;
     }
 
     if ((ev->event != ev->window) && (ev->event != root || !ev->send_event))
     {
-        TRACE ("handleUnmapNotify (): Event ignored\n");
+        TRACE ("handleUnmapNotify (): Event ignored");
         return;
     }
 
@@ -995,7 +1017,7 @@ handleConfigureNotify (XConfigureEvent * ev)
 static inline void
 handleConfigureRequest (XConfigureRequestEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
     XWindowChanges wc;
     XEvent otherEvent;
 
@@ -1003,9 +1025,9 @@ handleConfigureRequest (XConfigureRequestEvent * ev)
     TRACE ("ConfigureRequest on window (0x%lx)", ev->window);
 
     /* Compress events - logic taken from kwin */
-    while (XCheckTypedWindowEvent (dpy, ev->window, ConfigureRequest,
-            &otherEvent))
+    while (XCheckTypedWindowEvent (dpy, ev->window, ConfigureRequest, &otherEvent))
     {
+        last_timestamp = stashEventTime (last_timestamp, &otherEvent);
         if (otherEvent.xconfigurerequest.value_mask == ev->value_mask)
         {
             ev = &otherEvent.xconfigurerequest;
@@ -1115,6 +1137,10 @@ handleConfigureRequest (XConfigureRequestEvent * ev)
             }
             constrained = TRUE;
         }
+        if (ev->value_mask & CWStackMode)
+        {
+	    clientPassGrabButton1 (NULL);
+	}
 #if 0
         /* Let's say that if the client performs a XRaiseWindow, we show the window if hidden */
         if ((ev->value_mask & CWStackMode) && (wc.stack_mode == Above))
@@ -1142,7 +1168,7 @@ handleConfigureRequest (XConfigureRequestEvent * ev)
 static inline void
 handleEnterNotify (XCrossingEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleEnterNotify");
 
@@ -1161,7 +1187,11 @@ handleEnterNotify (XCrossingEvent * ev)
         TRACE ("EnterNotify window is \"%s\"", c->name);
         if ((c->type != WINDOW_DOCK) && (c->type != WINDOW_DESKTOP))
         {
-            clientSetFocus (c, TRUE, FALSE);
+            clientSetFocus (c, FOCUS_SORT);
+            if (!(params.raise_on_click))
+            {
+                clientPassGrabButton1 (c);
+            }
         }
     }
 }
@@ -1188,17 +1218,17 @@ handleLeaveNotify (XCrossingEvent * ev)
 static inline void
 handleFocusIn (XFocusChangeEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleFocusIn");
-    TRACE ("handleFocusOut (0x%lx) mode = %s",
+    TRACE ("handleFocusIn (0x%lx) mode = %s",
                 ev->window,
                 (ev->mode == NotifyNormal) ?
                 "NotifyNormal" :
                 (ev->mode == NotifyWhileGrabbed) ?
                 "NotifyWhileGrabbed" :
                 "(unknown)");
-    TRACE ("handleFocusOut (0x%lx) detail = %s",
+    TRACE ("handleFocusIn (0x%lx) detail = %s",
                 ev->window,
                 (ev->detail == NotifyAncestor) ?
                 "NotifyAncestor" :
@@ -1218,8 +1248,27 @@ handleFocusIn (XFocusChangeEvent * ev)
                 "NotifyDetailNone" :
                 "(unknown)");
 
-    if ((ev->mode == NotifyGrab) || (ev->mode == NotifyUngrab)
-        || (ev->detail > NotifyNonlinearVirtual))
+    if ((ev->window == root) && (ev->mode == NotifyNormal) && 
+        (ev->detail == NotifyDetailNone))
+    {
+        /* Handle focus transition to root (means that an unknown
+           window has vanished and the focus is returned to the root
+         */
+         
+        c = clientGetFocus ();
+        if (c)
+        {
+            clientSetFocus (c, FOCUS_FORCE);
+        }
+        else
+        {
+            clientSetFocus (NULL, FOCUS_NONE);
+            clientPassGrabButton1 (NULL);
+        }
+        return;
+    }
+    else if ((ev->mode == NotifyGrab) || (ev->mode == NotifyUngrab) ||
+             (ev->detail > NotifyNonlinearVirtual))
     {
         /* We're not interested in such notifications */
         return;
@@ -1230,7 +1279,7 @@ handleFocusIn (XFocusChangeEvent * ev)
     if (c)
     {
         TRACE ("focus set to \"%s\" (0x%lx)", c->name, c->window);
-        clientUpdateFocus (c);
+        clientUpdateFocus (c, FOCUS_NONE);
         if (params.raise_on_focus && !params.click_to_focus)
         {
             reset_timeout ();
@@ -1241,7 +1290,8 @@ handleFocusIn (XFocusChangeEvent * ev)
 static inline void
 handleFocusOut (XFocusChangeEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
+    
     TRACE ("entering handleFocusOut");
     TRACE ("handleFocusOut (0x%lx) mode = %s",
                 ev->window,
@@ -1276,10 +1326,11 @@ handleFocusOut (XFocusChangeEvent * ev)
     {
         c = clientGetFromWindow (ev->window, WINDOW);
         TRACE ("FocusOut on window (0x%lx)", ev->window);
-        if (c && (c == clientGetFocus ()))
+        if ((c) && (c == clientGetFocus ()))
         {
             TRACE ("focus lost from \"%s\" (0x%lx)", c->name, c->window);
-            clientUpdateFocus (NULL);
+            clientUpdateFocus (NULL, FOCUS_NONE);
+            clientPassGrabButton1 (NULL);
             /* Clear timeout */
             clear_timeout ();
         }
@@ -1289,7 +1340,7 @@ handleFocusOut (XFocusChangeEvent * ev)
 static inline void
 handlePropertyNotify (XPropertyEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handlePropertyNotify");
 
@@ -1394,7 +1445,7 @@ handlePropertyNotify (XPropertyEvent * ev)
 static inline void
 handleClientMessage (XClientMessageEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleClientMessage");
 
@@ -1502,10 +1553,11 @@ handleClientMessage (XClientMessageEvent * ev)
             TRACE
                 ("client \"%s\" (0x%lx) has received a net_active_window event",
                 c->name, c->window);
-            workspaceSwitch (c->win_workspace, NULL);
+            clientSetWorkspace (c, workspace, TRUE);
             clientShow (c, TRUE);
             clientRaise (c);
-            clientSetFocus (c, TRUE, FALSE);
+            clientSetFocus (c, FOCUS_SORT);
+            clientPassGrabButton1 (c);
         }
     }
     else
@@ -1549,7 +1601,7 @@ handleClientMessage (XClientMessageEvent * ev)
 static inline void
 handleShape (XShapeEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleShape");
 
@@ -1563,7 +1615,7 @@ handleShape (XShapeEvent * ev)
 static inline void
 handleColormapNotify (XColormapEvent * ev)
 {
-    Client *c;
+    Client *c = NULL;
 
     TRACE ("entering handleColormapNotify");
 
@@ -1583,6 +1635,7 @@ handleEvent (XEvent * ev)
     TRACE ("entering handleEvent");
 
     sn_process_event (ev);
+    last_timestamp = stashEventTime (last_timestamp, ev);
     switch (ev->type)
     {
         case MotionNotify:
